==> polars-0.14.5.tar.gz: polars-0.14.5/local_dependencies/polars-arrow/src/kernels/ewm/average.rs <==
use arrow::array::PrimitiveArray;
use arrow::types::NativeType;
use num::Float;

use crate::utils::CustomIterTools;

// See:
// https://github.com/pola-rs/polars/issues/2148
// https://stackoverflow.com/a/51392341/6717054

pub fn ewm_mean<T>(
    xs: &PrimitiveArray<T>,
    alpha: T,
    adjust: bool,
    min_periods: usize,
) -> PrimitiveArray<T>
where
    T: Float + NativeType,
{
    let mut denom = T::zero();
    let one_sub_alpha = T::one() - alpha;
    let mut non_null_cnt = 0usize;

    let mut opt_ewma = None;

    xs.iter()
        .map(|opt_x| {
            if let Some(&x) = opt_x {
                non_null_cnt += 1;

                let prev_ewma = opt_ewma.unwrap_or(x);

                let value = if adjust {
                    let numer = prev_ewma * denom * one_sub_alpha + x;
                    denom = T::one() + one_sub_alpha * denom;
                    numer / denom
                } else {
                    x * alpha + prev_ewma * one_sub_alpha
                };
                opt_ewma = Some(value);
            }
            match non_null_cnt < min_periods {
                true => None,
                false => opt_ewma,
            }
        })
        .collect_trusted()
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_ewm_mean_without_null() {
        let xs = PrimitiveArray::from([Some(1.0f32), Some(2.0f32), Some(3.0f32)]);

        for adjust in [false, true] {
            let result = ewm_mean(&xs, 0.5, adjust, 0);

            let expected = match adjust {
                false => PrimitiveArray::from([Some(1.0f32), Some(1.5f32), Some(2.25f32)]),
                true => PrimitiveArray::from([
                    Some(1.0f32),
                    Some(1.6666666666666667f32),
                    Some(2.4285714285714284f32),
                ]),
            };
            assert_eq!(result, expected);
        }
    }

    #[test]
    fn test_ewm_mean_with_null() {
        let xs = PrimitiveArray::from([Some(1.0f32), None, Some(1.0f32), Some(1.0f32)]);
        let result = ewm_mean(&xs, 0.5, false, 2);
        let expected = PrimitiveArray::from([None, None, Some(1.0f32), Some(1.0f32)]);
        assert_eq!(result, expected);

        let xs = PrimitiveArray::from([None, None, Some(1.0f32), Some(1.0f32)]);
        let result = ewm_mean(&xs, 0.5, false, 1);
        let expected = PrimitiveArray::from([None, None, Some(1.0f32), Some(1.0f32)]);
        assert_eq!(result, expected);

        let xs = PrimitiveArray::from([
            Some(2.0f32),
            Some(3.0f32),
            Some(5.0f32),
            Some(7.0f32),
            None,
            None,
            None,
            Some(4.0f32),
        ]);
        let result = ewm_mean(&xs, 0.5, false, 0);
        let expected = PrimitiveArray::from([
            Some(2.0f32),
            Some(2.5f32),
            Some(3.75f32),
            Some(5.375f32),
            Some(5.375f32),
            Some(5.375f32),
            Some(5.375f32),
            Some(4.6875f32),
        ]);
        assert_eq!(result, expected);

        let xs = PrimitiveArray::from([
            None,
            None,
            Some(5.0f32),
            Some(7.0f32),
            None,
            Some(2.0f32),
            Some(1.0f32),
            Some(4.0f32),
        ]);
        let unadjusted_result = ewm_mean(&xs, 0.5, false, 1);
        let unadjusted_expected = PrimitiveArray::from([
            None,
            None,
            Some(5.0f32),
            Some(6.0f32),
            Some(6.0f32),
            Some(4.0f32),
            Some(2.5f32),
            Some(3.25f32),
        ]);
        assert_eq!(unadjusted_result, unadjusted_expected);
        let adjusted_result = ewm_mean(&xs, 0.5, true, 1);
        let adjusted_expected = PrimitiveArray::from([
            None,
            None,
            Some(5.0f32),
            Some(6.333333333333333f32),
            Some(6.333333333333333f32),
            Some(3.857142857142857f32),
            Some(2.3333333333333335f32),
            Some(3.193548387096774f32),
        ]);
        assert_eq!(adjusted_result, adjusted_expected);

        let xs = PrimitiveArray::from([
            None,
            Some(1.0f32),
            Some(5.0f32),
            Some(7.0f32),
            None,
            Some(2.0f32),
            Some(1.0f32),
            Some(4.0f32),
        ]);
        let result = ewm_mean(&xs, 0.5, true, 1);
        let expected = PrimitiveArray::from([
            None,
            Some(1.0f32),
            Some(3.6666666666666665f32),
            Some(5.571428571428571f32),
            Some(5.571428571428571f32),
            Some(3.6666666666666665f32),
            Some(2.2903225806451615f32),
            Some(3.1587301587301586f32),
        ]);
        assert_eq!(result, expected);
    }
}

==> polars-0.14.12.tar.gz: polars-0.14.12/local_dependencies/polars-arrow/src/kernels/ewm/average.rs <==
use std::ops::AddAssign;

use arrow::array::PrimitiveArray;
use arrow::types::NativeType;
use num::Float;

use crate::utils::CustomIterTools;

pub fn ewm_mean<T>(
    xs: &PrimitiveArray<T>,
    alpha: T,
    adjust: bool,
    min_periods: usize,
) -> PrimitiveArray<T>
where
    T: Float + NativeType + AddAssign,
{
    let one_sub_alpha = T::one() - alpha;

    let mut opt_mean = None;
    let mut non_null_cnt = 0usize;

    let wgt = alpha;
    let mut wgt_sum = if adjust { T::zero() } else { T::one() };

    xs.iter()
        .map(|opt_x| {
            if let Some(&x) = opt_x {
                non_null_cnt += 1;

                let prev_mean = opt_mean.unwrap_or(x);

                wgt_sum = one_sub_alpha * wgt_sum + wgt;

                let curr_mean = prev_mean + (x - prev_mean) * wgt / wgt_sum;

                opt_mean = Some(curr_mean);
            }
            match non_null_cnt < min_periods {
                true => None,
                false => opt_mean,
            }
        })
        .collect_trusted()
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn test_ewm_mean_without_null() {
        let xs = PrimitiveArray::from([Some(1.0f32), Some(2.0f32), Some(3.0f32)]);

        for adjust in [false, true] {
            let result = ewm_mean(&xs, 0.5, adjust, 0);

            let expected = match adjust {
                false => PrimitiveArray::from([Some(1.0f32), Some(1.5f32), Some(2.25f32)]),
                true => PrimitiveArray::from([
                    Some(1.0f32),
                    Some(1.6666667f32), // <-- pandas: 1.66666667
                    Some(2.42857143),
                ]),
            };
            assert_eq!(result, expected);
        }
    }

    #[test]
    fn test_ewm_mean_with_null() {
        let xs = PrimitiveArray::from([Some(1.0f32), None, Some(1.0f32), Some(1.0f32)]);
        let result = ewm_mean(&xs, 0.5, false, 2);
        let expected = PrimitiveArray::from([None, None, Some(1.0f32), Some(1.0f32)]);
        assert_eq!(result, expected);

        let xs = PrimitiveArray::from([None, None, Some(1.0f32), Some(1.0f32)]);
        let result = ewm_mean(&xs, 0.5, false, 1);
        let expected = PrimitiveArray::from([None, None, Some(1.0f32), Some(1.0f32)]);
        assert_eq!(result, expected);

        let xs = PrimitiveArray::from([
            Some(2.0f32),
            Some(3.0f32),
            Some(5.0f32),
            Some(7.0f32),
            None,
            None,
            None,
            Some(4.0f32),
        ]);
        let result = ewm_mean(&xs, 0.5, false, 0);
        let expected = PrimitiveArray::from([
            Some(2.0f32),
            Some(2.5f32),
            Some(3.75f32),
            Some(5.375f32),
            Some(5.375f32),
            Some(5.375f32),
            Some(5.375f32),
            Some(4.6875f32),
        ]);
        assert_eq!(result, expected);

        let xs = PrimitiveArray::from([
            None,
            None,
            Some(5.0f32),
            Some(7.0f32),
            None,
            Some(2.0f32),
            Some(1.0f32),
            Some(4.0f32),
        ]);
        let unadjusted_result = ewm_mean(&xs, 0.5, false, 1);
        let unadjusted_expected = PrimitiveArray::from([
            None,
            None,
            Some(5.0f32),
            Some(6.0f32),
            Some(6.0f32),
            Some(4.0f32),
            Some(2.5f32),
            Some(3.25f32),
        ]);
        assert_eq!(unadjusted_result, unadjusted_expected);
        let adjusted_result = ewm_mean(&xs, 0.5, true, 1);
        let adjusted_expected = PrimitiveArray::from([
            None,
            None,
            Some(5.0f32),
            Some(6.33333333f32),
            Some(6.33333333f32),
            Some(3.85714286f32),
            Some(2.3333335f32), // <-- pandas: 2.33333333
            Some(3.19354839f32),
        ]);
        assert_eq!(adjusted_result, adjusted_expected);

        let xs = PrimitiveArray::from([
            None,
            Some(1.0f32),
            Some(5.0f32),
            Some(7.0f32),
            None,
            Some(2.0f32),
            Some(1.0f32),
            Some(4.0f32),
        ]);
        let result = ewm_mean(&xs, 0.5, true, 1);
        let expected = PrimitiveArray::from([
            None,
            Some(1.0f32),
            Some(3.66666667f32),
            Some(5.57142857f32),
            Some(5.57142857f32),
            Some(3.66666667),
            Some(2.2903228f32), // <-- pandas: 2.29032258
            Some(3.15873016f32),
        ]);
        assert_eq!(result, expected);
    }
}
